from enum import Enum

class Integration:

    class Types(Enum):
        CSV = 'CSV'
        EXCEL = 'EXCEL'
        GOOGLESPREADSHEET = 'GOOGLE SPREADSHEET'

    class Modes(Enum):
        READ = 'READ'
        WRITE = 'WRITE'

class WebScraper:

    class Functions:
        SUBSTRING = 'substring'
        EXPLODE = 'explode'
        REPLACE = 'join'
        JOIN = 'join'

    class Types:
        PERCENTAGE = 'PERCENTAGE'
        BOOL = 'BOOL'
        MONEY = 'MONEY'
        NUMBER = 'NUMBER'
        INTEGER = 'INTEGER'
        GEO = 'GEO'
        CHAR = 'CHAR'
        URL = 'URL'
        LIST = 'LIST'
        DATE = 'DATE'
        JSON = 'JSON'
        STRING = 'STRING'

    class Properties:
        TEXT = 'text()'
        HTML = 'html()'

class Requests:

    class Methods(Enum):
        POST = 'POST'
        GET = 'GET'
        PUT = 'PUT'
    
    class ResponseTypes(Enum):
        TEXT = 'TEXT'
        JSON = 'JSON'


from datetime import datetime, date, time
from sqlalchemy import create_engine, event
from sqlalchemy.orm import scoped_session, sessionmaker, aliased, mapper, relationship
from sqlalchemy.dialects import mysql
from sqlalchemy import exists, desc, text, and_
from sqlalchemy.inspection import inspect
from scrapeanything.database.models import Model
import os
import traceback

from scrapeanything.utils.types import Types
from scrapeanything.utils.utils import Utils
from scrapeanything.utils.config import Config
from scrapeanything.utils.log import Log

class Repository:

    def __init__(self, config: Config) -> None:
        self.config = config

        is_debug = self.config.get(section='COMMANDS', key='debug')

        username = self.config.get(section='DATABASE', key='username')
        password = self.config.get(section='DATABASE', key='password')
        host = self.config.get(section='DATABASE', key='host')
        port = self.config.get(section='DATABASE', key='port')
        database = self.config.get(section='DATABASE', key='database')
        self.engine = create_engine(f'mysql+pymysql://{username}:{password}@{host}:{port}/{database}', echo=is_debug) # , isolation_level='READ UNCOMMITTED')
        Session = sessionmaker(bind=self.engine)
        self.session = scoped_session(Session)
  
    def save(self, entity: any) -> any:
        '''
        Description: save a collection of models or a single model
        Arguments: <data> - collection or single model
        Returns: last id inserted or updated
        '''
        try:
            data = entity

            model_from_database = self.load(entity=data)
            model = self.update(entity_from_database=model_from_database, entity=data)

            if model not in self.session:
                self.session.merge(model)
            # else:
            #     self.session.merge(model)

            # self.session.commit()

            entity_type = type(model)
            Log.trace('Saved entity {0} with id {1}', entity_type, model.id)

            return model
        except Exception as e:
            Log.error(f'An error ocurred: {e}')
            Log.error(traceback.format_exc())


    def update(self, entity_from_database: any, entity: any) -> any:
        '''
        Description: updates an entity from its version on the database
        Arguments: 
        - <entity_from_database> - entity loaded from the database
        - <entity> - entity built in memory
        Returns: updated entity (holds Id from the database saved entity)
        '''

        if entity_from_database is None:
            return entity

        mapper = inspect(entity)
        for attr in mapper.attrs:
            if attr.key == 'id' or attr.key == 'created_at' or attr.key == 'updated_at':
                continue
            # elif attr.key.endswith('s'):
            #     continue
            elif attr.key.endswith('id'):
                continue

            value = getattr(entity, attr.key)
            if Types.is_primitive(value=value) and getattr(entity_from_database, attr.key) != value:
                setattr(entity_from_database, attr.key, value)

        if entity_from_database is not None:
            setattr(entity_from_database, 'updated_at', datetime.now())

        return entity_from_database

    def load(self, entity: any):
        '''
        Description:
        Arguments:
        Returns:
        '''

        entity_type = type(entity)

        id = getattr(entity, 'id')
        if id is not None:
            condition = getattr(entity_type, 'id') == entity.id
            return self.session.query(entity_type).filter(condition).first()

        filters = self.get_keys(entity=entity)
        from_database = self.first(entity_type, filters)
        return from_database

    def first(self, entity_type: any, filters: list):
        '''
        Description: returns the first element found given some filters
        Arguments:
        Returns:
        '''

        query = self.session.query(entity_type)

        aliases = {}
        for attr, value in filters.items():
            # condition is of type object.subobject.property = value
            if attr.find('.') > -1:

                entities = attr.split('.')[:-1] # from object.subobject1.subobject2.property take object.subobject1.subobject2
                attribute = attr.split('.')[-1] # from object.subobject1.subobject2.property last attribute is property

                parent_entity_type = entity_type
                for i, sub_entity in enumerate(entities): # from object.subobject1.subobject2.property get only object.subobject1.subobject2
                    # subobject<i>
                    sub_entity_type = Types.to_class(module_name='database.models', class_name=self.get_sub_entity_type(entity=parent_entity_type, prop=sub_entity))
                    aliases[sub_entity] = aliased(sub_entity_type)
                    # join
                    query = query.join(aliases[sub_entity], Utils.find(parent_entity_type, sub_entity))

                    # joins have to be executed on [object, subobject1]; [suboject1, suboject2]; etc.
                    parent_entity_type = sub_entity_type

                # filter
                # load an object by similarity
                if self.config.get(section='DATABASE', key='load_filter', default='equal') == 'like':
                    condition = getattr(aliases[sub_entity], attribute).like(f'%{value}%')
                else:
                    condition = getattr(aliases[sub_entity], attribute) == value

                query = query.filter(condition)

            # condition is of type object.property = value
            elif attr.find('.') == -1:
                if self.config.get(section='DATABASE', key='load_filter', default='equal') == 'like':
                    condition = getattr(entity_type, attr).like(f'%{value}%')
                else:
                    condition = getattr(entity_type, attr) == value

                # add condition to the query 
                query = query.filter(condition)

        entity = query.first()
        return entity

    def all(self, entity_type):
        '''
        Description:
        Arguments:
        Returns:
        '''

        entity_type = Types.to_class(module_name='database.models', class_name=entity_type)
        entities = self.session.query(entity_type).all()
        return entities

    def get_keys(self, entity: any):
        '''
        Description: Returns list of columns and their values to determine uniquely a given entity (defined in models.py)
        Arguments:
        Returns:
        '''

        def normalize_keys(keys: list):
            normalized_keys = []

            for key in keys:
                if key.find('*') > -1:
                    entity = key.split('.')[-2] # take the last entity before *
                    entity_type = entity.split('_')[0].capitalize()
                    entity_type = Types.to_class(module_name='database.models', class_name=entity_type)

                    for token in entity_type.__keys__.split(','):
                        token = token.strip()
                        normalized_key = key.replace('*', token)
                        normalized_keys += normalize_keys([ normalized_key ])
                else:
                    normalized_keys.append(key)

            return normalized_keys

        def replace_star_in_key(keys: str):
            # supports __keys__ = match.*

            response = []

            for key in keys.split(','):
                key = key.strip()
                if key.find('*') > -1:
                    ref_entity = key.strip().split('.')[-2] # another_entity
                    ref_entity_type = ref_entity.split('_')[0].capitalize().strip() if ref_entity.find('_') != -1 else ref_entity.capitalize().strip() # extract entity name # TODO: use self.get_sub_entity_type()?
                    ref_entity_type = Types.to_class(module_name='database.models', class_name=ref_entity_type.capitalize().strip())
                    exploded_keys = ref_entity_type.__keys__ # __keys__ of another_entity

                    # exploded_keys are taken from child entity (i.e. match.* => championship.name, team_home.name, team_away.name).
                    # Prepende at each reference property the source object (i.e. match.championship.name, match.team_home.name, match.team_away.name)
                    for exploded_key in exploded_keys.split(','):
                        exploded_key = exploded_key.strip()
                        if exploded_key.find('*') > -1:
                            exploded_key = f'{ref_entity}.{exploded_key.strip()}'
                            response.append(replace_star_in_key(keys=exploded_key))
                        else:
                            response.append(f'{ref_entity}.{exploded_key.strip()}')
                else:
                    response.append(key)

            return response
        
        filters = {}

        # for <entity> there could be some columns that refer to __keys__ of another entity (i.e. other_entity.another_entity.*)
        __keys__ = normalize_keys(keys=[ key.strip() for key in entity.__keys__.split(',') ])

        # get all keys listed comma separated
        for key in __keys__:
            filters[key] = eval(f'entity.{key}') # Utils.find(entity, key) # { key: value }

        return filters

    def get_raw_query(self, q: any):
        '''
        Description: transforms SqlAlchemy query into t-sql query (It can be used for debugging purposes)
        Arguments: <q> - SqlAlchemy query (from self.session(<q>))
        Returns: t-sql query string version
        '''

        query = q.statement.compile(dialect=mysql.dialect())
        params = query.params # get query parameters and values

        query_string = str(q.as_scalar()) # get query in t-sql format. It contains constants as query parameters (i.e. SELECT * FROM table WHERE value = :value)
        for param, value in params.items():

             # replace query parameters with real values
            if type(value) is str:
                value = f"'{value}'"
            elif type(value) is date:
                value = f"'{datetime.strftime(value, '%Y-%m-%d')}'"
            elif type(value) is time:
                value = f"'{datetime.strftime(value, '%H-%M-%S')}'"
            elif type(value) is datetime:
                value = f"'{datetime.strftime(value, '%Y-%m-%d %H-%M-%S')}'"

            query_string = query_string.replace(f':{param}', str(value))

        query_string = query_string.replace('\n', '') # the query contains some "\n". Let's remove them
        query_string = query_string[1:-1] # remove initial "(" and last ")" form query_string

        return query_string

    def get_sub_entity_type(self, entity: any, prop: str):
        '''
        Description: <TODO>
        Arguments: <TODO>
        Returns: <TODO>
        '''

        try:
            return getattr(entity, prop).property.argument
        except:
            key = getattr(entity, prop).property.key
            if key == 'team_home' or 'team_away':
                return 'Team'
            if key == 'match':
                return 'Match'
            if key == 'bet':
                return 'Bet'

    def query(self, entity_type):
        return self.session.query(entity_type)

    def sql(self, query: str, parameters: dict=None, entity_type: any=None) -> any:
        '''
        Executes sql query. Make sure you use :parameter for parameters
        '''
        q = text(query)
        result = self.session.execute(q, parameters).mappings().all()
        return result

    def assign(self, entity, prop, value):
        loaded_value = self.load(entity=value)
        value = self.update(entity_from_database=loaded_value, entity=value)
        if type(value) is list:
            for v in value:
                if v not in self.session:
                    value.is_dirty = True
                setattr(entity, prop, value)
        else:
            if value not in self.session:
                value.is_dirty = True
            setattr(entity, prop, value)

        return entity

    def rollback(self):
        self.session.rollback()

    def commit(self):
        self.session.commit()

    def close(self):
        self.session.close()
        self.engine.dispose()


import dateutil.parser
import datetime
import pytz
import tzlocal
import time

class Dates:

    @staticmethod
    def diff_months(d1: datetime, d2: datetime) -> int:
        return (d1.year - d2.year) * 12 + d1.month - d2.month

    @staticmethod
    def get_current_timezone() -> str:
        return tzlocal.get_localzone().zone

    @staticmethod
    def parse_datetime(datetime: datetime) -> datetime:
        return dateutil.parser.parse(datetime).replace(tzinfo=None)

    @staticmethod
    def now(type: str=None) -> datetime:
        date = datetime.datetime.now()

        if type == 'DATE':
            return date.date()
        elif type == 'TIME':
            return date.time()
        else:
            return date

    @staticmethod
    def subtract(date: datetime, value: int, unit_of_measure: str) -> datetime:
        if unit_of_measure == 'SECONDS':
            delta = datetime.timedelta(seconds=value)
        elif unit_of_measure == 'MINUTES':
            delta = datetime.timedelta(minutes=value)
        elif unit_of_measure == 'HOURS':
            delta = datetime.timedelta(hours=value)
        elif unit_of_measure == 'DAYS':
            delta = datetime.timedelta(days=value)

        return date - delta

    @staticmethod
    def change_timezone(datetime: datetime, from_timezone_code: str='Europe/Rome', to_timezone_code: str='US/Pacific') -> datetime:
        # create both timezone objects
        old_timezone = pytz.timezone(from_timezone_code)
        new_timezone = pytz.timezone(to_timezone_code)

        return old_timezone.localize(datetime).astimezone(new_timezone)

    @staticmethod
    def get_timezone() -> str:
        return time.tzname

    @staticmethod
    def difference(datetime1: datetime, datetime2: datetime, measure: str) -> float:
        if measure == 'SECONDS':
            return (datetime2 - datetime1).total_seconds()
        elif measure == 'MINUTES':
            return (datetime2 - datetime1).total_seconds() / 60
        elif measure == 'HOURS':
            return (datetime2 - datetime1).total_seconds() / 3600
        elif measure == 'DAYS':
            return (datetime2 - datetime1).total_seconds() / 86400


import requests
from scrapeanything.database.repository import Repository
from scrapeanything.utils.config import Config
from scrapeanything.utils.service import Service

class TelegramBot(Service):

    def __init__(self, config: Config=None, repository: Repository=None) -> None:
        super().__init__(config, repository)

        self.TOKEN = self.config.get(section='NOTIFIERS', key='TELEGRAM_CHATID')
        self.CHAT_ID = self.config.get(section='NOTIFIERS', key='TELEGRAM_TOKEN')
        # self.TOKEN = '6023559959:AAHYuMGqqqAtO4I5D3DSX5THtVLGaYgsuTI'
        # self.CHAT_ID = '-1001845468851'

    def get_chat_id(self) -> str:
        url = f"https://api.telegram.org/bot{self.TOKEN}/getUpdates"
        print(requests.get(url).json())

    def send_message(self, message: str) -> None:
        url = f"https://api.telegram.org/bot{self.TOKEN}/sendMessage?chat_id={self.CHAT_ID}&text={message}"
        print(requests.get(url).json()) # this sends the message

    def send_image(self, filename: str, caption: str=None) -> None:
        url = f'https://api.telegram.org/bot{self.TOKEN}/sendPhoto'
        files = {'photo': open(f'{filename}.png', 'rb')}
        data = {'chat_id': self.CHAT_ID, 'caption': caption }
        print(requests.post(url, files=files, data=data))


from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from webdriver_manager.chrome import ChromeDriverManager

import time
import platform
import random
from scrapeanything.scraper.scraper import Scraper
from scrapeanything.utils.config import Config
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import Select
from selenium.common.exceptions import NoSuchElementException        
from selenium.common.exceptions import StaleElementReferenceException
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.remote.remote_connection import LOGGER
import undetected_chromedriver as uc

import logging
import base64

class Selenium(Scraper):

    def __init__(self, config: Config, headless: bool=True, fullscreen: bool=False, window: dict={}, user_dir: str=None, disable_javascript: bool=False, proxy_address: str=None):
        url = config.get(section='SELENIUM', key='url')

        LOGGER.setLevel(logging.WARNING)
        
        options = Options()
        options.add_argument('--disable-blink-features=AutomationControlled')        
        options.add_argument('--disable-gpu')
        options.add_argument('disable-infobars')
        options.add_argument('--disable-extensions')
        options.add_argument('disable-infobars')
        options.add_experimental_option('useAutomationExtension', False)
        options.add_experimental_option('excludeSwitches', ['enable-automation'])
        options.add_argument('--no-sandbox')
        options.add_argument('--disable-dev-shm-usage')

        user_agents = [
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/92.0.4515.159 Safari/537.36",
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.114 Safari/537.36",
            "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:89.0) Gecko/20100101 Firefox/89.0",
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:89.0) Gecko/20100101 Firefox/89.0",
        ]
        options.add_argument(f"user-agent={random.choice(user_agents)}")        

        window_size = window['size'] if 'size' in window else []
        if len(window_size) == 0:
            options.add_argument('--start-maximized')
        else:
            self.driver.set_window_position(x=0, y=0)
            self.driver.set_window_size(width=window_size[0], height=window_size[1])

        if fullscreen is True:
            options.add_argument('--kiosk')

        if headless is True:
            options.add_argument('--headless')

        if proxy_address is not None:
            options.add_argument(f'--proxy-server={proxy_address}')

        if url is not None:
            self.driver = webdriver.Remote(command_executor=url, options=options)
        elif url is None:
            service = Service(ChromeDriverManager().install()) # Use webdriver_manager to automatically download and manage the correct version of chromedriver
            self.driver = webdriver.Chrome(service=service, options=options) # Initialize the WebDriver

        self.driver.execute_script('Object.defineProperty(navigator, "webdriver", {get: () => undefined})')

        super().__init__()


    # def __init__(self, config: Config, headless: bool=True, fullscreen: bool=False, window: dict={}, user_dir: str=None, disable_javascript: bool=False, proxy_address: str=None):
        
    #     url = config.get(section='SELENIUM', key='url')

    #     if user_dir is None or url is not None:
    #         options = webdriver.ChromeOptions()
    #     else:
    #         options = uc.ChromeOptions()

    #         # another way to set profile is the below (which takes precedence if both variants are used
    #         options.add_argument(f'--user-data-dir={user_dir}')
    #         # just some options passing in to skip annoying popups
    #         options.add_argument('--no-first-run --no-service-autorun --password-store=basic')            

    #     #region options
    #     if fullscreen is True:
    #         options.add_argument('--kiosk')

    #     if proxy_address is not None:
    #         options.add_argument(f'--proxy-server={proxy_address}')

    #     if disable_javascript is True:
    #         options.add_experimental_option('prefs', { 'profile.managed_default_content_settings.javascript': 2 })

    #     options.add_argument('--no-sandbox')
    #     options.add_argument('--log-level=3')        
    #     options.add_argument('--disable-extensions')
    #     options.add_argument('--disable-dev-shm-usage')

    #     if user_dir is None:
    #         options.add_argument('start-maximized')
    #         options.add_argument('disable-infobars')
    #         options.add_argument('--disable-extensions')
    #         options.add_argument('disable-infobars')
    #         options.add_experimental_option('useAutomationExtension', False)
    #         options.add_experimental_option('excludeSwitches', ['enable-automation'])

    #     window_size = window['size'] if 'size' in window else []

    #     if len(window_size) == 0 and user_dir is not None:
    #         options.add_argument('--start-maximized')

    #     if headless is True:
    #         options.add_argument('headless')

    #     LOGGER.setLevel(logging.WARNING)

    #     # options.add_argument('--disable-blink-features=AutomationControlled')
    #     # options.add_argument('user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/74.0.3729.169 Safari/537.36')

    #     #endregion options

    #     if user_dir is not None:
    #         self.driver = uc.Chrome(options=options)
    #     elif url is not None:
    #         self.driver = webdriver.Chrome(options=options)
    #     elif url is None:
    #         service = Service()
    #         self.driver = webdriver.Chrome(service=service, options=options)

    #     if len(window_size) > 0:
    #         self.driver.set_window_position(x=0, y=0)
    #         self.driver.set_window_size(width=window_size[0], height=window_size[1])

    #     super().__init__()

    def on_wget(self, url):
        self.driver.get(url)
        return self.driver

    def on_xPath(self, path, element=None, timeout=0, is_error: bool=True):
        return (self.driver if element is None else element).find_elements(by=By.XPATH, value=path)

    def on_exists(self, path, element, timeout=0, is_error: bool=False):
        element = self.driver if element is None else element

        if timeout > 0:
            try:
                ignored_exceptions = (NoSuchElementException, StaleElementReferenceException)
                element = WebDriverWait(driver=element, timeout=timeout, ignored_exceptions=ignored_exceptions).until(
                    EC.presence_of_element_located((By.XPATH, path)))
                
                if isinstance(element, list): return len(element) > 0
                else: return element is not None

            except Exception as ex:
                if is_error is True: raise NoSuchElementException(msg=f'Element with xpath {path} was not found')
                else: return False
        else:
            try:
                if is_error is False:
                    return len(element.find_elements(by=By.XPATH, value=path)) > 0
                elif is_error is True and len(element.find_elements(by=By.XPATH, value=path)) == 0:
                    raise NoSuchElementException(msg=f'Element with xpath {path} was not found')

            except NoSuchElementException as ex:
                if is_error is True:
                    raise ex
                else: return False
            except StaleElementReferenceException:
                raise StaleElementReferenceException()

    def on_get_text(self, element):
        return element.text

    def on_get_html(self, element):
        return element.get_attribute('innerHTML')

    def on_get_attribute(self, element, prop):
        return element.get_attribute(prop)

    def click_all(self, path: str, element: any=None, timeout=0, is_error: bool=True) -> None:
        elements = self.xPath(path=path, element=element, timeout=timeout, is_error=is_error)
        if elements is not None:

            try:
                element.click()
            except Exception as e:
                try:
                    self.driver.execute_script('arguments[0].click();', element)
                except:
                    try:
                        actions = ActionChains(self.driver)
                        actions.click(element).perform()
                    except:
                        return None

            return element
        else:
            return None

    def on_hover(self, path: str, element: any, is_error: bool=True, timeout: int=0) -> None:
        element = self.xPath(element=element, path=path, is_error=is_error, timeout=timeout)
        ActionChains(self.driver).move_to_element(element).perform()

    def on_click_and_hold(self, path: str, seconds: float, element: any=None, timeout: int=0, is_error: bool=True) -> None:
        element = self.xPath(path=path, timeout=timeout, is_error=is_error)
        action = ActionChains(self.driver)
        action.click_and_hold(element).perform()
        time.sleep(seconds) # sleep for N seconds
        action.release().perform()

    def on_click(self, path: str, element: any=None, timeout: int=0, pos: int=0, is_error: bool=True) -> any:
        if element is None and path is not None:
            element = self.xPath(path=path, element=element, timeout=timeout, pos=pos, is_error=is_error)
        
        if element is not None:
            if type(element) is list:
                if len(element) > 0:
                    element = element[pos]
                else:
                    return None
            try:
                element.click()
            except Exception as e:
                try:
                    self.driver.execute_script('arguments[0].click();', element)
                except Exception as e:
                    try:
                        actions = ActionChains(self.driver)
                        actions.click(element).perform()
                    except Exception as e:
                        return None

                return element
        else:
            return None

    def on_back(self):
        self.driver.back()

    def on_get_current_url(self):
        return self.driver.current_url

    def on_enter_text(self, path: str, text: str, clear: bool=False, element: any=None, timeout: int=0, is_error: bool=True):
        element = self.on_xPath(path=path, element=element, timeout=timeout, is_error=is_error)

        if element is not None:
            if type(element) is list:
                element = element[0]

            if clear is True:
                element.send_keys(Keys.CONTROL + 'a')
                element.send_keys(Keys.BACKSPACE)

            element.send_keys(text)
            return True
        else:
            return False

    def on_solve_captcha(self, path: str):
        pass

    def on_login(self, username_text=None, username=None, password_text=None, password=None, is_error: bool=True):
        self.on_enter_text(username_text, username, is_error=is_error)
        self.on_enter_text(password_text, password, is_error=is_error)

        self.on_click('//button[@class="sign-in-form__submit-button"]')

    def on_search(self, path=None, text=None, timeout: int=0, is_error: bool=True):
        self.on_enter_text(path=path, text=text, timeout=timeout, is_error=is_error)
        return self.on_enter_text(path=path, text=Keys.RETURN, timeout=timeout, is_error=is_error)

    def on_scroll_to_bottom(self):
        self.driver.execute_script('window.scrollTo(0, document.body.scrollHeight);')

    def on_get_scroll_top(self):
        return self.driver.execute_script('return document.documentElement.scrollTop;')

    def on_get_scroll_bottom(self):
        return self.driver.execute_script('return document.body.scrollHeight')

    def on_select(self, path: str, option: str, element: any=None) -> None:
        element = WebDriverWait(self.driver if element is None else element, 10).until(
            EC.presence_of_element_located((By.XPATH, path))
        )

        if element is not None:
            Select(element).select_by_visible_text(option)

    def on_get_image_from_canvas(self, path, local_path, element):
        image_filename = f'{local_path}/image.png'

        # get the base64 representation of the canvas image (the part substring(21) is for removing the padding 'data:image/png;base64')
        base64_image = self.driver.execute_script(f"return document.querySelector('{path}').toDataURL('image/png').substring(21);")

        # decode the base64 image
        output_image = base64.b64decode(base64_image)

        # save to the output image
        with open(image_filename, 'wb') as f:
            f.write(output_image)

        return image_filename

    def on_switch_to(self, element: any) -> None:
        if element == 'default':
            self.driver.switch_to.default_content()
        else:
            self.driver.switch_to.frame(element)

    def on_screenshot(self, file: str) -> None:
        self.driver.get_screenshot_as_file(file)

    def on_freeze(self) -> None:
        self.driver.execute_cdp_cmd('Emulation.setScriptExecutionDisabled', { 'value': True })

    def on_unfreeze(self) -> None:
        self.driver.execute_cdp_cmd('Emulation.setScriptExecutionDisabled', { 'value': False })

    def on_close(self):
        self.driver.quit()


from configparser import ConfigParser
import re

class Config:

    def __init__(self, path: str):
        self.config = ConfigParser()
        # read config.ini file
        self.config.read(path)

    def set(self, section: str, key: str, value: str) -> None:
        if section in self.config and key in self.config[section]:
            self.config[section][key] = value

    def get(self, section: str, key: str, default: any=None):
        # get config value
        if section in self.config and key in self.config[section]:
            response = self.config[section][key]

            if response == 'True': return True
            elif response == 'False': return False
            elif bool(re.search('^(?=.)([+-]?([0-9]*)(\.([0-9]+))?)$', response)) is True and '.' in response: return float(response)
            elif bool(re.search('^(?=.)([+-]?([0-9]*)(\.([0-9]+))?)$', response)) is True and '.' not in response: return int(response)
            else: return response

        elif (section in self.config and key not in self.config[section]) or (section not in self.config):
            if default is not None: return default
            else: return None
        else: raise Exception('Configuration not found!')